import platform
import subprocess
from ..constants import *
from ..layers import *
from ..utils import *
from ..materials import *
import json
import gdsfactory as gf
from copy import deepcopy
# from time import time
import datetime
from math import cos, pi, sin
import os
import numpy as np

from sortedcontainers import SortedDict, SortedSet
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def setup(path, c, study, nres, center_wavelength,
          bbox_layer=BBOX_LAYER,
          zlims=None, core_layer=LAYER.WG,
          port_margin="auto",
          runs=[],  sources=[],
          layer_stack=SOI, materials=dict(),
          default_material='SiO2',
          exclude_layers=[], Courant=None,
          gpu=None, dtype=np.float32,
          plot=False, framerate=None,
          magic="", wd=os.path.join(os.getcwd(), "runs"), name=None,
          Ttrans=None,
          approx_2D_mode=False):
    materials = {**MATERIALS, **materials}
    prob = {
        'nres': nres,
        'center_wavelength': center_wavelength,
        'name': name,
        'path': path,
    }
    if approx_2D_mode:
        N = 2
        prob["approx_2D_mode"] = approx_2D_mode
    else:
        N = 3
        prob["approx_2D_mode"] = None
    prob["class"] = "pic"
    prob["Ttrans"] = Ttrans
    prob["dtype"] = str(dtype)
    prob["timestamp"] = datetime.datetime.now().isoformat(
        timespec="seconds").replace(":", "-")
    prob["magic"] = magic
    prob["framerate"] = framerate

    gpu_backend = gpu
    # if gpu_backend:s
    prob["gpu_backend"] = gpu_backend

    ports = {
        p.name: {
            "center": (np.array(p.center)/1e0).tolist(),
            "normal": [cos(p.orientation/180*pi), sin(p.orientation/180*pi)],
            "tangent": [-sin(p.orientation/180*pi), cos(p.orientation/180*pi)],
            'width': p.width/1e0,
        }
        for p in c.get_ports_list(prefix="o")
    }
    for (k, v) in ports.items():
        z, n = [0, 0, 1], [*v['normal'], 0]
        t = np.cross(z, n).tolist()
        v['frame'] = [t, z, n]
        v["dimensions"] = [v['width']],
    prob["ports"] = ports
    mode_solutions = []

    d = get_layers(layer_stack, core_layer)[0]
    hcore = d.thickness
    zcore = d.zmin

    zmargin = 4*hcore
    zcenter = zcore+hcore/2

    h = hcore+2*zmargin
    zmin = zcore-zmargin
    zmax = zmin+h

    source_ports = []
    nonsource_ports = []
    for p in c.ports:
        is_source = False
        for run in runs:
            for port in run["sources"]:
                if port == p.name:
                    is_source = True
        if is_source:
            source_ports.append(p.name)
        else:
            nonsource_ports.append(p.name)
#
    port_width = max([p.width/1e0 for p in c.ports])
    ps = portsides(c)
    xmargin = ymargin = 2*port_width

    source_port_margin = 2 * port_width if N == 2 else 6*port_width

    port_margin = center_wavelength/nres
    margins = []
    for p in ps:
        if set(p).intersection(source_ports):
            margins.append(source_port_margin+port_margin)
        elif set(p).intersection(nonsource_ports):
            margins.append(port_margin)
        else:
            assert not p
            margins.append(xmargin)

    #
    modexmargin = .6*xmargin
    modezmargin = .6*zmargin
    wmode = port_width+2*modexmargin
    hmode = hcore+2*modezmargin
    zmode = zcore-modezmargin

    prob["hcore"] = hcore
    prob["zcenter"] = zcenter
    prob["zmin"] = zmin
    prob["zmax"] = zmax
    prob["zcore"] = zcore
    # prob["L"] = [l, w, h]

    _c = gf.Component()
    kwargs = dict()
    for (orientation, side, length, p) in zip([0, 90, 180, 270], ["right", "top", "left", "bottom"], margins, ps):
        if p:
            c = gf.components.extend_ports(
                c, None, length, orientation=orientation)
        else:
            kwargs[side] = length
    _c << gf.components.bbox(component=c, layer=bbox_layer, **kwargs)
    # _c = gf.Component()
    _c << c
    c = _c
    # c.plot()
    # c.show()

    layers = set(c.layers)-set(exclude_layers)

    TEMP = os.path.join(path, "temp")
    os.makedirs(TEMP, exist_ok=True)
    BODIES = os.path.join(path, 'bodies')
    os.makedirs(BODIES, exist_ok=True)

    layer_stack_info = material_voxelate(
        c,  zmin, zmax, layers, layer_stack, BODIES)
    dir = os.path.dirname(os.path.realpath(__file__))
    print(dir)
    fn = os.path.join(dir, "solvemodes.py")
    assert os.path.exists(fn)
    assert os.path.exists(TEMP)
    if platform.system() == "Windows":
        os.system(f"copy /Y {fn} {TEMP}")
    else:
        subprocess.run(["cp", fn, TEMP])
    prob["layer_stack"] = layer_stack_info
    prob["study"] = study
    prob["materials"] = materials

    prob["N"] = N

    prob["hmode"] = hmode
    prob["wmode"] = wmode
    prob["zmode"] = zmode
    wmode = port_width+2*modexmargin
    wavelengths = []
    # _c = add_bbox(c, layer=bbox_layer, nonport_margin=margin)
    for run in runs:
        for k, v in list(run["sources"].items())+list(run["monitors"].items()):
            p = ports[k]
            # v['dimensions'] = p['dimensions'][0]+[hmode]
            v['dimensions'] = [wmode, hmode]
            v['frame'] = p['frame']
            for center_wavelength in v["wavelength_mode_numbers"]:
                wavelengths.append(center_wavelength)
        for k, v in run['sources'].items():
            p = ports[k]
            ct = np.array(p['center'])
            n = np.array(p['normal'])
            v["center"] = (ct+n*source_port_margin).tolist()
            v['center'] += [zcenter]
        for k, v in run['monitors'].items():
            p = ports[k]
            v['center'] = copy.deepcopy(p['center'])
            v['center'] += [zcenter]

    wavelengths = sorted(set(wavelengths))

    prob["mode_solutions"] = mode_solutions
    prob["runs"] = runs
    bbox = c.bbox_np().tolist()
    bbox[0].append(zmin)
    bbox[1].append(zmax)

    prob['bbox'] = bbox
    # prob['epdefault'] = materials[layer_stack['default']['material']]['epsilon']
    prob['epdefault'] = materials[default_material]['epsilon']

    prob["Courant"] = Courant
    if not os.path.exists(path):
        os.makedirs(path)
    # prob["wavelengths"] = wavelengths
    c.write_gds(os.path.join(path, "component.gds"))
    prob["mode_solutions"] = mode_solutions
    # prob["path_length"] = 2*(l+r+lwg)
    return prob


def port_name(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        return s
    return f"o{s}"


def port_number(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        s = s[1:]
    return int(s)


def mode_number(port):
    l = str(port).split("@")
    return 0 if len(l) == 1 else int(l[1])


def unpack_sparam_key(k):
    o, i = k.split(",")
    po, pi = port_name(o), port_name(i)
    mo, mi = mode_number(o), mode_number(i)
    return po, mo, pi, mi


def long_sparam_key(k):
    po, mo, pi, mi = unpack_sparam_key(k)
    return f"{po}@{mo},{pi}@{mi}"



import os
# import pymeshfix
import json
from statistics import median
from .constants import *
import gdsfactory as gf
from gdsfactory.cross_section import Section
from .constants import *
from .materials import *
try:
    from IPython.display import display
except ImportError:
    pass
# import pyvista as pv

# from .picToGDS import main

from math import ceil, cos, pi, sin, tan
import matplotlib.pyplot as plt
import numpy as np
from gdsfactory.generic_tech import LAYER_STACK, LAYER
import copy
import shutil
import trimesh
# from gdsfactory import LAYER_VIEWS

tol = .001


def get(c, i):
    try:
        return c[i]
    except:
        try:
            return getattr(c, i)
        except:
            return None


def arange(a, b, d):
    ret = np.linspace(a, b, round((b-a)/(d))+1).tolist()
    return ret


def trim(x, dx):
    return round(x/dx)*dx


def extend(endpoints, wm):
    v = endpoints[1]-endpoints[0]
    v = v/np.linalg.norm(v)
    return [
        (endpoints[0]-wm*v).tolist(), (endpoints[1]+wm*v).tolist()]


# def round(x):
#     r = round(x)
#     if abs(x-r) > .1 and r == 0:
#         raise ValueError(f"round: {x} -> {r}")

#     return r


def portsides(c):
    ports = c.ports
    bbox = c.bbox_np()
    res = [[], [], [], []]
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    for p in ports:
        x, y = np.array(p.center)/1e0

        if abs(x - xmin0) < tol:
            res[2].append(p.name)
        if abs(x - xmax0) < tol:
            res[0].append(p.name)
        if abs(y - ymin0) < tol:
            res[3].append(p.name)
        if abs(y - ymax0) < tol:
            res[1].append(p.name)
    return res


def add_bbox(c, layer, nonport_margin=0):  # , dx=None):
    bbox = c.bbox_np()
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    l = xmax0-xmin0
    w = ymax0-ymin0

    # l = dx*np.ceil((xmax0-xmin0)/dx)
    # w = dx*np.ceil((ymax0-ymin0)/dx)

    # if dx is not None:
    #     if nonport_margin is None:
    #         nonport_margin = dx
    # if nonport_margin is None:
    #     nonport_margin = 0
    margin = nonport_margin
    xmin, ymin, xmax, ymax = xmin0-margin, ymin0 - \
        margin, xmin0+l+margin, ymin0+w+margin

    for p in c.ports:
        # p = c.ports[k]
        x, y = np.array(p.center)/1e0
        if abs(x - xmin0) < tol:
            xmin = x
        if abs(x - xmax0) < tol:
            xmax = x
        if abs(y - ymin0) < tol:
            ymin = y
        if abs(y - ymax0) < tol:
            ymax = y
    p = [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
    _c = gf.Component()
    _c << c

    if type(layer[0]) is int:
        layer = [layer]
    for layer in layer:
        layer = tuple(layer)
        _c.add_polygon(p, layer=layer)
    for port in c.ports:
        _c.add_port(name=port.name, port=port)
    return _c

# def pic2gds(fileName, sizeOfTheCell, layerNum=1, isDither=False, scale=1):
    main(fileName, sizeOfTheCell, layerNum, isDither, scale)
    return "image.bmp", "image.gds"


def finish(c, name):
    c.add_label(name, position=c.bbox_np()[1])


def normal_from_orientation(orientation):
    return [cos(orientation/180*pi), sin(orientation/180*pi)]


def material_voxelate(c,  zmin, zmax, layers, layer_stack, path):
    stacks = sum([[[v.mesh_order, v.material, tuple(layer), k]
                 for k, v in get_layers(layer_stack, layer, withkey=True)] for layer in layers], [])
    c.flatten()
    stacks = sorted(stacks, key=lambda x: -x[0])
    layer_stack_info = dict()
    # raise NotImplementedError("This is a stub")
    lb, ub = c.bbox_np()
    # bbox = [[**lb, zmin], [**ub, zmax]]
    bbox = [[lb[0], lb[1], zmin], [ub[0], ub[1], zmax]]
    layers = [x[2] for x in stacks]
    for i, stack in enumerate(stacks):
        order = stack[0]
        m = stack[1]
        l1, l2 = layer = stack[2]
        k = stack[3]

        _layer_stack = copy.deepcopy(layer_stack)
        get(_layer_stack, 'layers').clear()

        d = copy.deepcopy(get(layer_stack, 'layers')[k])
        if d.zmin <= zmax and d.bounds[1] >= zmin:
            get(_layer_stack, 'layers')[k] = d
            d.zmin = max(zmin, d.zmin)
            d.thickness = min(zmax-d.zmin, d.thickness)
            # _d.bounds = (_d.zmin, _d.zmin+_d.thickness)
            # origin = (c.extract([layer]).bbox_np()-c.bbox_np())[0].tolist()

            mesh = c.to_3d(
                layer_stack=_layer_stack,
                layer_views=LAYER_VIEWS,
                exclude_layers=set(layers)-{layer})
            OBJ = os.path.join(path, f'{order}_{m}_{k}.obj')
            trimesh.exchange.export.export_mesh(mesh, OBJ, 'obj')
            # pymeshfix.clean_from_file(OBJ, OBJ)
            # repair_obj_open3d(OBJ, OBJ)

            # STL = os.path.abspath(os.path.join(path, f"{k}.stl"))
            # gf.export.to_stl(c, STL, layer_stack=_layer_stack)

            # STL = os.path.join(path, f'{k}_{l1}_{l2}.stl')
            # pymeshfix.clean_from_file(STL, STL)
            # STL1 = os.path.join(path, f'{order}_{m}_{k}.stl')
            # os.replace(STL, STL1)

            # mesh = pv.read(STL)
            # im = stl_to_array(mesh, dl*unit, bbox)
            # np.save(os.path.join(path, f'{k}.npy'), im)

            layer_stack_info[k] = {
                "layer": (l1, l2),
                "zmin": d.zmin,
                "thickness": d.thickness,
                "material": matname(m),
                "mesh_order": stack[0],
                # "origin": origin,
            }
    return layer_stack_info


def get_layers(layer_stack, layer, withkey=False):
    r = []
    d = get(layer_stack, 'layers').items()

    for k, x in d:
        l = get(x, 'layer')
        if l is not None:
            t = get(l, 'layer')
            if t is not None and tuple(t) == tuple(layer):
                if withkey:
                    x = k, x
                r.append(x)
    if r:
        return r

    for k, x in d:
        l = get(x, 'derived_layer')
        if l is not None:
            t = get(l, 'layer')
            if t is not None and tuple(t) == tuple(layer):
                if withkey:
                    x = k, x
                r.append(x)
    return r


def wavelength_range(center, bandwidth, length=3):
    f1 = 1/(center+bandwidth/2)
    f2 = 1/(center-bandwidth/2)
    hw = (f2-f1)/2
    f1 = 1/center-hw
    f2 = 1/center+hw
    return sorted([1/x for x in np.linspace(f1, f2, length).tolist()])


def adjust_wavelengths(wavelengths, center_wavelength=None, wl_res=.01):
    if len(wavelengths) == 1:
        return wavelengths, wavelengths[0], 1
    wavelengths = sorted(set(wavelengths))
    wavelengths0 = copy.deepcopy(wavelengths)
    if center_wavelength is None:
        center_wavelength = median(wavelengths)
    freqs0 = [center_wavelength/w for w in reversed(wavelengths)]
    nresfreq = round(1/min(np.diff([0]+freqs0)))

    while True:
        freqs = [round(f*nresfreq)/nresfreq for f in freqs0]
        wavelengths = [center_wavelength/f for f in reversed(freqs)]
        if max([abs(w-w0)/center_wavelength for w, w0 in zip(wavelengths, wavelengths0)]) <= wl_res:
            break
        nresfreq += 1
    print(
        f"wavelengths has been adjusted within `wl_res` to facilitate simulation:\n{wavelengths}")
    return wavelengths, center_wavelength, nresfreq


def save_problem(prob, path):
    path = os.path.abspath(path)
    bson_data = json.dumps(prob)
    # prob["component"] = c0

    path = prob["path"]
    if not os.path.exists(path):
        os.makedirs(path)
        #   compiling julia code...
        #   """)
    prob_path = os.path.join(path, "problem.json")
    with open(prob_path, "w") as f:
        # Write the BSON data to the file
        f.write(bson_data)
    print("using simulation folder", path)



import gdsfactory as gf
from gdsfactory.generic_tech.layer_map import LAYER
from gdsfactory.technology import LogicalLayer, LayerStack, LayerLevel
import gdsfactory as gf
from gdsfactory.generic_tech import LAYER, LAYER_STACK
from gdsfactory.generic_tech.get_klayout_pyxs import get_klayout_pyxs
from gdsfactory.technology import LayerLevel, LayerStack, LayerViews, LayerMap
from gdsfactory.generic_tech import get_generic_pdk

from gdsfactory.config import PATH
from gdsfactory.technology import (
    LayerLevel,
    LayerStack,
    LayerView,
    LayerViews,
    LayerMap,
)

Layer = tuple[int, int]

# gf.generic_tech.


class MyLayerMap(LayerMap):
    WG: Layer = (1, 0)
    WGCLAD: Layer = (111, 0)
    BOX: Layer = (112, 0)
    PORT: Layer = (1, 10)
    DESIGN: Layer = (200, 0)
    GUESS: Layer = (201, 0)


LAYER = MyLayerMap

nm = 1e-3
thickness_wg = 220 * nm
thickness_slab_deep_etch = 90 * nm
thickness_slab_shallow_etch = 150 * nm
box_thickness = 0.5
thickness_clad = 0.5

layer_core = LogicalLayer(layer=LAYER.WG)
layer_clad = LogicalLayer(layer=LAYER.WGCLAD)
layer_box = LogicalLayer(layer=LAYER.BOX)
CANVAS_LAYER = LogicalLayer(layer=CANVAS_LAYER)

LAYER_STACK.layers.update(dict(
    core=LayerLevel(
        layer=layer_core,
        thickness=thickness_wg,
        zmin=0,
        material="Si",
        mesh_order=2,
    ),
    box=LayerLevel(
        layer=layer_box,
        thickness=box_thickness,
        zmin=-box_thickness,
        material="SiO2",
        mesh_order=9,
    ),
    clad=LayerLevel(
        layer=layer_clad,
        zmin=0.0,
        material="SiO2",
        thickness=thickness_clad,
        mesh_order=10,
    ),
    design=LayerLevel(
        layer=CANVAS_LAYER,
        zmin=0.0,
        material=None,
        thickness=.0001,
        mesh_order=100,
    ),
))

pdk = get_generic_pdk()
pdk.activate()
LAYER_VIEWS = pdk.layer_views
LAYER_VIEWS.layer_views["WGCLAD"].visible = True
LAYER_VIEWS.layer_views["canvas"] = LayerView(layer=CANVAS_LAYER, visible=True)
# for l in LAYER_STACK.layers.values():
#     print(l.layer)
#     print(l.layer.layer)



# import gplugins.luminescent as gl
from gdsfactory.generic_tech import *
from .runs_utils import *
from .utils import *
from .constants import *
from .materials import *
from .sol import *

from .pic.gcells import *
from .pic.setup import *
from .pic.inverse_design import *
from .pic.sparams import *

from .gen.setup import *
print("running luminescent python frontend")


# __all__ = [
#     "pic_design_problem",
#     "sparams_problem",
#     "solve",
#     # "apply_design",
#     "load",
#     "finetune",
#     "load_problem",
#     # "make_training_movie",
#     # "make_simulation_movie",
#     "mimo",
# ]



# import dill
from pprint import pprint
import os
import json
# import cv2
import numpy as np
from .pic.inverse_design import *
from .pic.sparams import *
from .utils import *
from .layers import *
from .constants import *


def lastrun(name="", wd=os.path.join(os.getcwd(), "runs"), study="",  **kwargs):
    if name:
        return os.path.join(wd, name)
    l = [os.path.join(wd, x) for x in os.listdir(wd)]
    l = [x for x in l if os.path.isfile(os.path.join(x, "problem.json"))]
    l = sorted(l, key=lambda x: os.path.getmtime(x), reverse=True)
    if study:
        for x in l:
            try:
                if json.loads(open(os.path.join(x, "problem.json"), "rb").read())["study"] == study:
                    return x
            except:
                pass
    return l[0]


def load_problem(path):
    path = os.path.abspath(path)
    print(f"loading problem from {path}")
    return json.loads(open(os.path.join(path, "problem.json"), "rb").read())


def make_simulation_movie(path, framerate=None, **kwargs):
    TEMP = os.path.join(path, "temp")
    FIELDS = os.path.join(TEMP, "fields")
    FRAMES = os.path.join(TEMP, "frames")
    MOVIE = os.path.join(path, "simulation_video.mp4")
    shutil.rmtree(FRAMES, ignore_errors=True)
    os.makedirs(FRAMES, exist_ok=True)
    shutil.rmtree(MOVIE, ignore_errors=True)

    if framerate is None:
        framerate = load_problem(path)["framerate"]

    g = np.load(os.path.join(path, "temp", 'g.npy')).T
    gmax = np.max(np.abs(g))

    dir = os.path.join(path, "temp", 'fields')
    umax = 0
    fns = sorted(os.listdir(dir), key=lambda x: float(x[0:-4]))
    for fn in fns:
        a = np.load(os.path.join(dir, fn))
        v = np.max(np.abs(a))
        if umax < v:
            umax = v

    for fn in fns:
        name = fn[0:-4]
        a = np.load(os.path.join(FIELDS, fn)).T
        fig, axs = plt.subplots(1, 2)
        axs[1].imshow(a, cmap='seismic', origin='lower',
                      vmin=-umax, vmax=umax)
        axs[0].imshow(-g, cmap='gray',
                      origin='lower', vmin=-gmax, vmax=0)
        plt.savefig(os.path.join(FRAMES, f"{name}.png"))
        plt.close(fig)

    fns = sorted(os.listdir(FRAMES), key=lambda x: float(x[0:-4]))
    frame = height = width = layers = video = None
    for i, fn in enumerate(fns):
        frame = cv2.imread(os.path.join(FRAMES, fn))
        if i == 0:
            height, width, layers = frame.shape
            video = cv2.VideoWriter(
                MOVIE, 0x7634706d, framerate, (width, height))

        video.write(frame)

    cv2.destroyAllWindows()
    video.release()


def make_training_movie(framerate=2, **kwargs):
    1
    # path = lastrun(**kwargs)

    # f = os.path.join(path, "checkpoints")
    # ckpts = sorted(os.listdir(f))
    # frame = cv2.imread(os.path.join(f, ckpts[0], "run_1.png"))
    # height, width, layers = frame.shape

    # video = cv2.VideoWriter(os.path.join(
    #     path, "training_video.mp4"), 0x7634706d, framerate, (width, height))

    # for ckpt in ckpts:
    #     video.write(cv2.imread(os.path.join(f, ckpt, "run_1.png")))

    # cv2.destroyAllWindows()
    # video.release()



from gdsfactory.technology import LogicalLayer, LayerLevel, LayerStack
from gdsfactory.generic_tech.layer_map import LAYER
import gdsfactory as gf
import copy
from gdsfactory.generic_tech import LAYER_STACK, LAYER

MATKEYS = {
    "si": "cSi",
    "Si": "cSi",

    "sio2": "SiO2",
    "sin": "SiN",
    "ge": "Ge",

}


def matname(k):
    if k in MATKEYS:
        return MATKEYS[k]
    return k.capitalize()


MATERIALS = {
    "cSi": {'epsilon': 3.48**2},
    "SiO2": {'epsilon': 1.44**2},
    "SiN": {'epsilon': 2.0**2},
    "Ge": {'epsilon': 4.0**2},
    "Si": {'epsilon': 3.48**2},
    'PEC': {'epsilon': 10000},
}

ks = copy.deepcopy(list(MATERIALS.keys()))
for k in ks:
    MATERIALS[k.lower()] = MATERIALS[k]


nm = 1e-3
thickness_wg = 220 * nm
thickness_slab_deep_etch = 90 * nm
thickness_slab_shallow_etch = 150 * nm

sidewall_angle_wg = 0
layer_core = LogicalLayer(layer=LAYER.WG)
layer_shallow_etch = LogicalLayer(layer=LAYER.SHALLOW_ETCH)
layer_deep_etch = LogicalLayer(layer=LAYER.DEEP_ETCH)

layers = {
    "core": LayerLevel(
        layer=LogicalLayer(layer=LAYER.WG),
        thickness=thickness_wg,
        zmin=0.0,
        material="si",
        mesh_order=2,
        # sidewall_angle=sidewall_angle_wg,
        # width_to_z=0.5,
    ),
    # "shallow_etch": LayerLevel(
    #     layer=LogicalLayer(layer=LAYER.SHALLOW_ETCH),
    #     thickness=thickness_wg - thickness_slab_shallow_etch,
    #     zmin=0.0,
    #     material="si",
    #     mesh_order=1,
    #     derived_layer=LogicalLayer(layer=LAYER.SLAB150),
    # ),
    # "deep_etch": LayerLevel(
    #     layer=LogicalLayer(layer=LAYER.DEEP_ETCH),
    #     thickness=thickness_wg - thickness_slab_deep_etch,
    #     zmin=0.0,
    #     material="si",
    #     mesh_order=1,
    #     derived_layer=LogicalLayer(layer=LAYER.SLAB90),
    # ),
    # "slab150": LayerLevel(
    #     layer=LogicalLayer(layer=LAYER.SLAB150),
    #     thickness=150e-3,
    #     zmin=0,
    #     material="si",
    #     mesh_order=3,
    # ),
    # "slab90": LayerLevel(
    #     layer=LogicalLayer(layer=LAYER.SLAB90),
    #     thickness=thickness_slab_deep_etch,
    #     zmin=0.0,
    #     material="si",
    #     mesh_order=2,
    # ),
}


SOI = LayerStack(layers=layers)
SOI.layers['default'] = {
    'material': 'SiO2'
}
BBOX_LAYER = (8888, 8888)



from copy import deepcopy
from gdsfactory.generic_tech import get_generic_pdk
from gdsfactory.technology import (
    LayerLevel,
    LayerStack,
    LayerView,
    LayerViews,
    LayerMap,
)
CANVAS_LAYER = (1000, 1)

pdk = get_generic_pdk()
pdk.activate()
LAYER_VIEWS = pdk.layer_views
LAYER_VIEWS.layer_views["WGCLAD"].visible = True
LAYER_VIEWS.layer_views["canvas"] = deepcopy(LAYER_VIEWS.layer_views["WGCLAD"])
LAYER_VIEWS.layer_views["canvas"].layer = CANVAS_LAYER
# LayerView(layer=CANVAS_LAYER, visible=True)

eps0 = 8.854187817e-12



# import dill
import skrf as rf
from pprint import pprint
import os
import subprocess
import json
import numpy as np
import requests
from .pic.inverse_design import *
from .pic.sparams import *
from .utils import *
from .layers import *
from .constants import *
from PIL import Image


from subprocess import Popen, PIPE
URL = "http://127.0.0.1:8975"


def start_fdtd_server(url=URL):
    try:
        r = requests.get(url)
    except:
        print("starting julia fdtd server on localhost...")
        cmd = ["julia", "-e", "using Luminescent; start_fdtd_server()"]
        proc = Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        with proc:
            for line in proc.stdout:
                print(str(line.decode().strip()), flush=True)
            err_message = proc.stderr.read().decode()
            print(err_message)


def solve(path, dev=False):
    path = os.path.abspath(path)
    # prob = load_problem(path)

    # prob["action"] = "solve"
    # r = requests.post(f"{url}/local", json=prob)
    1
    # print(f"""
    #       using simulation folder {path}
    #       starting julia process...
    #       """)
    # start_time = time.time()

    def run(cmd):
        print("="*40)
        print()
        proc = Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # proc.wait()
        with proc:
            for line in proc.stdout:
                print(str(line.decode().strip()), flush=True)
            err_message = proc.stderr.read().decode()
            print(err_message)
    # if dev:
    #     env = r'using Pkg;Pkg.activate(raw"c:\Users\pxshe\OneDrive\Desktop\beans\Luminescent.jl\luminescent");'
    # else:
    #     env = '0;'
    # cmd = ["lumi", path]
    # try:
    run(["Luminescent", path])
    # except:
    # run(["julia", "-e", f'println(Base.active_project())'])
    # print("no binaries found - starting julia session to compile - will alternate between execution and JIT compilation - will take 3 mins before simulation starts.\nYou can take a break and come back :) ...")

    # prob = json.loads(open(os.path.join(path, "problem.json"), "rb").read())
    # a = ['julia', '-e', ]
    # gpu_backend = prob["gpu_backend"]
    # _class = prob["class"]
    # if gpu_backend == "CUDA":
    #     array = "cu"
    #     pkgs = ",CUDA"
    # else:
    #     array = "Array"
    #     pkgs = ""

    # b = [f'using Luminescent{pkgs};{_class}run(raw"{path}",{array})']
    # run(a+b)

    # with Popen(cmd,  stdout=PIPE, stderr=PIPE) as p:
    #     if p.stderr is not None:
    #         for line in p.stderr:
    #             print(line, flush=True)
    # exit_code = p.poll()
    # subprocess.run()
    # print(f"julia simulation took {time.time()-start_time} seconds")
    # print(f"images and results saved in {path}")
    # sol = load(path=path)
    # return sol


def load_sparams(sparams):
    if "re" in list(sparams.values())[0]:
        return {k: v["re"]+1j*v["im"] for k, v in sparams.items()}
    return {center_wavelength: {k: (v["re"]+1j*v["im"])
                                for k, v in d.items()} for center_wavelength, d in sparams.items()}


def load(path, show=True):
    path = os.path.abspath(path)
    print(f"loading solution from {path}")
    prob = json.loads(open(os.path.join(path, "problem.json"), "rb").read())
    p = os.path.join(path, "solution.json")
    sol = json.loads(open(p).read())

    if prob['class'] == 'gen':
        S = np.load(os.path.join(path, "S.npy"))
        frequencies = rf.Frequency.from_f(sol['fs'], unit='GHz')
        ntwk = rf.Network(frequency=frequencies, s=S)
        ntwk.write_touchstone(os.path.join(path, 'S.s2p'))
        sol.update({"S": S, "network": ntwk})
    elif prob['class'] == 'pic':
        # sol = json.loads(p, "rb").read())["sol"]
        sol["S"] = load_sparams(sol["S"])
        sol["component"] = gf.import_gds(os.path.join(path, "component.gds"))
        if prob["study"] == "sparams":
            pass
        elif prob["study"] == "inverse_design":
            # l = [np.array(d) for d in sol["optimized_canvases"]]
            # sol["optimized_canvases"] = l
            # print(
            #     f"loading optimized design regions at resolution {sol['dl']}")
            # for i, a in enumerate(l):
            #     path = f"optimized_canvas_{i+1}.png"
            #     Image.fromarray(np.flip(np.uint8((1-a)) * 255, 0),
            #                     'L').save(os.path.join(path, name))
            # pic2gds(os.path.join(
            #     path, name), sol["dx"])
            # c = apply_design(sol["component"],  sol)
            # sol["optimized_component"] = copy.deepcopy(c)
            # sol["optimized_component"] = c
            # c.write_gds(os.path.join(path, "optimized_component.gds"))
            0

        if show:
            p = os.path.join(path, f"sim.png")
            if os.path.exists(p):
                img = Image.open(p)
                img.show()
                try:
                    display(img)
                except:
                    pass

            _sol = {k: sol[k] for k in ["T", "dB", "S",
                                        "phasors", "total_power_T", "total_power_dB"]}
            # _sol["optimized_canvases"] = "[[...]]"
            pprint(_sol)

    if show:
        for fn in os.listdir(path):
            if fn.endswith(".png") or fn.endswith(".jpg"):
                fn = os.path.join(path, fn)
                img = Image.open(fn)
                img.show()
                try:
                    display(img)
                except:
                    pass

    return sol


def finetune(path, iters, **kwargs):
    fn = os.path.join(path, "problem.json")
    prob = json.loads(open(fn, "rb").read())
    prob["iters"] = iters
    prob["path"] = path
    prob["restart"] = False
    prob = {**prob, **kwargs}
    with open(fn, 'w') as f:
        json.dump(prob, f)
    return solve(path)

    # PROB = os.path.join(path, "problem.json")
    # prob = json.loads(open(PROB, "rb").read())
    # prob["iters"] = iters
    # prob["path"] = path
    # prob["restart"] = False
    # prob = {**prob, **kwargs}
    # json.dump(prob, open(PROB, "w"))
    # return solve(path)
